/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package SchedulesRecovery;

import AbsenceSchedules.AbsenceSchedulesArray;
import java.util.*;
import ScheduleUtil.*;

/**
 *
 * @author frabe
 */
public class RecoveredSchedulesArray extends SchedulesArray{
    
    public AbsenceSchedulesArray absenceSchedulesArray;
    
    public RecoveredSchedulesArray(AbsenceSchedulesArray absenceSchedulesArray){
        super(absenceSchedulesArray);
        this.absenceSchedulesArray = absenceSchedulesArray;
        recoverAbsenceScheduleOptimally();
    }
    
    private void recoverAbsenceScheduleOptimally() {
        boolean[][] currentRecoveredScheduleOfPartTimeEmployees = this.getDeepCopyEmployeesSchedules(this.absenceSchedulesArray.partTimeSchedules);
        boolean[][] currentRecoveredScheduleOfFullTimeEmployees = this.getDeepCopyEmployeesSchedules(this.absenceSchedulesArray.fullTimeSchedules);      
       
        int remainingRecoveringAction = this.absenceSchedulesArray.numberOfRecoveringActionsToPerform;
       
        boolean[][] alreadyRecoveredWorkPeriodByPartTimeEmployees = new boolean[this.getNumberPartTimeEmployee()][remainingRecoveringAction];
        boolean[][] alreadyRecoveredWorkPeriodByFullTimeEmployees = new boolean[this.getNumberFullTimeEmployee()][remainingRecoveringAction];
        for (int i = 0 ; i < this.getNumberPartTimeEmployee() ; i++) {
            Arrays.fill(alreadyRecoveredWorkPeriodByPartTimeEmployees[i], Boolean.FALSE);
        }
        for (int i = 0 ; i < this.getNumberFullTimeEmployee() ; i++) {
            Arrays.fill(alreadyRecoveredWorkPeriodByFullTimeEmployees[i], Boolean.FALSE);
        }

        RecoveringActionComparator recoveringActionCostComparator = new RecoveringActionComparator();
        ScheduleState currentScheduleState = new ScheduleState(currentRecoveredScheduleOfPartTimeEmployees, currentRecoveredScheduleOfFullTimeEmployees, 
                                                                remainingRecoveringAction);

        Stack< ScheduleState > scheduleStatesNodes = new Stack<>();
        scheduleStatesNodes.push(currentScheduleState);
        RecoveringAction currentRecoveringAction;
        
        boolean isSolutionFound;
        isSolutionFound = remainingRecoveringAction == 0;
        // Fouilles en profondeur en privilegiant les actions avec cout minimum. On marque les noeuds dejà visites pour avoir des fouilles differentes
        while ( !isSolutionFound ) {
            PriorityQueue<RecoveringAction> recoveringActionsOrderedByCost =
                    getPossibleRecoveringActions(currentRecoveredScheduleOfPartTimeEmployees, currentRecoveredScheduleOfFullTimeEmployees, 
                    alreadyRecoveredWorkPeriodByPartTimeEmployees, alreadyRecoveredWorkPeriodByFullTimeEmployees, remainingRecoveringAction, recoveringActionCostComparator);
            if (recoveringActionsOrderedByCost.size() > 0){
                // On prend l'action avec un cout minimal
                currentRecoveringAction = recoveringActionsOrderedByCost.poll();
                
                boolean[][] recoveredScheduleOfPartTimeEmployeesAfterAction = this.getDeepCopyEmployeesSchedules(currentRecoveredScheduleOfPartTimeEmployees);
                boolean[][] recoveredScheduleOfFullTimeEmployeesAfterAction = this.getDeepCopyEmployeesSchedules(currentRecoveredScheduleOfFullTimeEmployees);
                
                if (currentRecoveringAction.getClass().getName().equals("SchedulesRecovery.RecoveringActionPartTimeEmployee")) {
                    recoveredScheduleOfPartTimeEmployeesAfterAction[currentRecoveringAction.employee][currentRecoveringAction.workPeriod] = true;    
                    alreadyRecoveredWorkPeriodByPartTimeEmployees[currentRecoveringAction.employee][remainingRecoveringAction-1] = true;
                } else if (currentRecoveringAction.getClass().getName().equals("SchedulesRecovery.RecoveringActionFullTimeEmployee")) {
                    recoveredScheduleOfFullTimeEmployeesAfterAction[currentRecoveringAction.employee][currentRecoveringAction.workPeriod] = true;   
                    alreadyRecoveredWorkPeriodByPartTimeEmployees[currentRecoveringAction.employee][remainingRecoveringAction-1] = true;
                }
                
                ScheduleState scheduleState = new ScheduleState(recoveredScheduleOfPartTimeEmployeesAfterAction, recoveredScheduleOfFullTimeEmployeesAfterAction, 
                                                   remainingRecoveringAction - 1);
                    
                scheduleStatesNodes.push(scheduleState);
            } else if (scheduleStatesNodes.size() == 1) { 
                //Si le sommet n'a plus d'enfants non-visite, la solution n'est pas trouvee.
                break;
            } else{
                // Lorsqu'on est bloque dans l'arbre de recherche, on revient au sommet de l'arbre.
                do{
                    scheduleStatesNodes.pop();
                } while(scheduleStatesNodes.size() > 1);

            }
            if (remainingRecoveringAction - 1 == 0){
                isSolutionFound = true;
            }
            if (scheduleStatesNodes.size() > 0) {
                currentScheduleState =  scheduleStatesNodes.peek();
                currentRecoveredScheduleOfPartTimeEmployees = currentScheduleState.currentRecoveredScheduleOfPartTimeEmployees;
                currentRecoveredScheduleOfFullTimeEmployees = currentScheduleState.currentRecoveredScheduleOfFullTimeEmployees;
                remainingRecoveringAction = currentScheduleState.remainingRecoveringAction;
            } 
        }

        if (isSolutionFound) {
            this.fullTimeSchedules = currentRecoveredScheduleOfFullTimeEmployees;
            this.partTimeSchedules = currentRecoveredScheduleOfPartTimeEmployees;
            this.totalScheduleCost = EmployeeCostCalculator.getFullRecoveredScheduleCost(this, myParametersInitialSchedules);
        } else {
            this.totalScheduleCost = Integer.MAX_VALUE;
        }
        
    } 
    
    private PriorityQueue<RecoveringAction> getPossibleRecoveringActions(boolean[][] currentPartTimeSchedule, boolean[][] currentFullTimeSchedule, 
            boolean[][] alreadyRecoveredWorkPeriodByPartTimeEmployees, boolean[][] alreadyRecoveredWorkPeriodByFullTimeEmployees, 
            int remainingRecoveringAction, Comparator recoveringActionCostComparator) {
        PriorityQueue<RecoveringAction> recoveringActionsOrderedByCost = new PriorityQueue<>(recoveringActionCostComparator);
        int workPeriod = findWorkPeriodToRecover(currentPartTimeSchedule, currentFullTimeSchedule);
        
        if (workPeriod < myParametersInitialSchedules.getWorkPeriodsPerSchedule()) {
            for (int partTimeEmployee = 0 ; partTimeEmployee < this.getNumberPartTimeEmployee() ; partTimeEmployee++ ) {
                if ( ValidationRecoveringAction.isPartTimeEmployeeAvailableForAbsenceRecovering(currentPartTimeSchedule, myParametersInitialSchedules, 
                            absenceSchedulesArray, partTimeEmployee, workPeriod) 
                        && !alreadyRecoveredWorkPeriodByPartTimeEmployees[partTimeEmployee][remainingRecoveringAction-1]){
                    RecoveringActionPartTimeEmployee recoveringAction = 
                            new RecoveringActionPartTimeEmployee(partTimeEmployee, workPeriod);
                    recoveringAction.calculateRecoveringActionCost (currentPartTimeSchedule, absenceSchedulesArray, myParametersInitialSchedules);
                    recoveringActionsOrderedByCost.add(recoveringAction);
                }
            }
            for (int fullTimeEmployee = 0 ; fullTimeEmployee < this.getNumberFullTimeEmployee() ; fullTimeEmployee++ ) {
                if ( ValidationRecoveringAction.isFullTimeEmployeeAvailableForAbsenceRecovering(currentFullTimeSchedule, myParametersInitialSchedules, 
                            absenceSchedulesArray, fullTimeEmployee, workPeriod) 
                        &&  !alreadyRecoveredWorkPeriodByFullTimeEmployees[fullTimeEmployee][remainingRecoveringAction-1] ){
                    RecoveringActionFullTimeEmployee recoveringAction = 
                            new RecoveringActionFullTimeEmployee(fullTimeEmployee, workPeriod);
                    recoveringAction.calculateRecoveringActionCost (currentFullTimeSchedule, absenceSchedulesArray, myParametersInitialSchedules);
                    recoveringActionsOrderedByCost.add(recoveringAction);
                }
            }
        }
        return recoveringActionsOrderedByCost;
    }

    private int findWorkPeriodToRecover(boolean[][] currentPartTimeSchedule, boolean[][] currentFullTimeSchedule){
        int workingEmployees;
        for (int workPeriod = 0 ; workPeriod < myParametersInitialSchedules.getWorkPeriodsPerSchedule() ; workPeriod++) {
            workingEmployees = 0;
            for (int partTimeEmployee = 0 ; partTimeEmployee < currentPartTimeSchedule.length ; partTimeEmployee++) {
                if ( currentPartTimeSchedule[partTimeEmployee][workPeriod] ) {
                    workingEmployees++;
                }
            }
            for (int fullTimeEmployee = 0 ; fullTimeEmployee < currentFullTimeSchedule.length ; fullTimeEmployee++) {
                if ( currentFullTimeSchedule[fullTimeEmployee][workPeriod] ) {
                    workingEmployees++;
                }
            }
            if (workingEmployees < this.myParametersInitialSchedules.getRequiredWorkForce(workPeriod) ) {
                return workPeriod;
            }
        }
        return myParametersInitialSchedules.getWorkPeriodsPerSchedule();
    }
}
